/*Top-Level Wrapper File to implement circuits with the FLEX10K70
on the Ultragizmo board. DO NOT change any of the pin definitions
unless you are completely sure of the consequences.
*/
module wrapper (address, data, codata, srama, sramdl, sramdh, codsync,
                codread, codwrite, sram1en, sram2en, sram1oe, sram2oe,
                sram1we, sram2we, sram1ud, sram2ud, sram1ld, sram2ld,
                br_sfpga, bg_sfpga, bgack_sfpga, clk, fc, berr, as, halt, reset,
                dtack, lds, uds, rw, irqsf, iacksf, hex0, hex1, hex2, hex3,
                sfpga_digital, sfpga_con40, sfpga_logic, sfpga_con60, led, pclk
              );

	input [23:1] address;
	inout [15:0] data;
	input [31:0] codata;
	input [17:0] srama;
	input [15:0] sramdl;
	input [15:0] sramdh;
	input codsync, codread, codwrite, sram1en, sram2en, sram1oe, sram2oe;
	input sram1we, sram2we, sram1ud, sram2ud, sram1ld, sram2ld;
	input br_sfpga, bg_sfpga, bgack_sfpga, clk;
	input [2:0] fc;
	input berr, as, halt, reset;
	inout dtack;
	input lds, uds, rw, iacksf;
	input [33:0] sfpga_digital;
	inout [31:0] sfpga_con40;
	input [17:0] sfpga_logic;
	input [51:0] sfpga_con60;
	input pclk; 

	output [7:0] hex0, hex1, hex2, hex3;
	output [15:0] led;
	output irqsf;

	wire reg_write, reg_read;
	wire start_out, stop_out, temp_out;
	wire [0:3] temp;
	wire [0:3] temp1;
	wire [0:3] temp2;
	wire led_drive;
	wire count_enb;
	wire new_clock;
	wire [19:0] counter_Q;

	communicate COM0 (address, as, rw, uds, lds, dtack, reg_write, reg_read);

	flipflop START (data[0], reg_write, 1'b1, 1'b0, start_out);
	flipflop STOP (data[1], reg_write, 1'b1, sfpga_con40[0], stop_out);

	hex7seg DHEX0 (counter_Q[3:0], hex3); 
	hex7seg DHEX1 (counter_Q[7:4], hex2);
	hex7seg DHEX2 (counter_Q[11:8], hex1);
	hex7seg DHEX3 (counter_Q[15:12], hex0);
	assign led[0] = new_clock;
	assign led[1] = new_clock;
	assign led[2] = new_clock;
	assign led[3] = new_clock;
	assign led[4] = new_clock;
	assign led[5] = new_clock;
	assign led[6] = new_clock;
	assign led[7] = new_clock;

	comb_logic(start_out, stop_out, irqsf, led_drive, count_enb);		
	leds(led_drive, sfpga_con40[19:12]);
	clockdiv (clk, new_clock);
	counter(new_clock, counter_Q, count_enb, reg_write); 
	tristatebuf(counter_Q, reg_read, data[15:0]);  

endmodule

module tristatebuf(in, control, out);
	input [15:0] in;
	input control;
	output [15:0] out;
	genvar i;

	generate
		for (i = 0; i < 16; i = i + 1)
		begin: tristates
			bufif1(out[i], in[i], control);
		end
	endgenerate
endmodule

module comb_logic(start_out, stop_out, inter, leds, count_enb);
	input start_out;
	input stop_out;
	output inter;
	output leds;
	output count_enb;

	assign leds = start_out & ~stop_out;
	assign inter = ~start_out | ~stop_out;
	assign count_enb = leds;	
endmodule

module leds(on, leds);
	input on;
	output [8:0] leds;

	reg [8:0] leds;

	always @(posedge on or negedge on)
	begin
		if (on)
			leds[8:0] = 8'hFF;
		else
			leds[8:0] = 8'h00;
	end
endmodule

module clockdiv(clk, new_clock);
	input clk;
	output new_clock;

	reg [19:0] counter;

	assign new_clock = counter[13];

	always @(posedge clk)
	begin
		counter = counter + 1;
	end
endmodule

module counter(clk, Q, enable, reset);
	input clk;
	input enable;
	input reset;
	output [15:0] Q;

	reg Q;

	always @(posedge clk or posedge reset)
		if (reset)
			Q <= 0;	
		else if (enable)
			Q <= Q + 1;
endmodule

module communicate(address, as, rw, uds, lds, dtack, reg_write, reg_read);

	input [23:1] address;
	input as, rw;
	input uds, lds;
	inout dtack;
	output reg_write, reg_read;

	wire reg_sel;
	wire ack, gnd;

	assign gnd = 0;
	bufif1(dtack, gnd, ack);

	   
	assign reg_sel = address[23] & (~ address[22]) & address[21]
	                   & address[20] & (~ as);
	   
	assign reg_write = reg_sel & (~ uds) & (~ lds) & (~ rw);
	assign reg_read = reg_sel & (~ uds) & (~ lds) & rw;

	assign ack = reg_read | reg_write;

endmodule

module flipflop(D, Clock, Resetn, Preset, Q);
	input D, Clock, Resetn, Preset;
	output Q;
	reg Q;
	
	always @(posedge Preset or negedge Resetn or posedge Clock)
	begin
		if (Preset)
			Q <= 1;
		else if (!Resetn)
			Q <= 0;
		else
			Q <= D;
	end
endmodule

