#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "ubi_dLinkList.h"

#define MAXDEGREE 24
#define ZERO 1
#define ONE 2
#define X 4
#define EMPTY 8
#define EPSILON 16

// We only need one degree size per invocation of the program.
static int degree;

// Note that a cube with the first element 0 indicates
// an empty cube, ie {0x0}*{111} = null set

// Just to make it clear
typedef unsigned char literal_t;

typedef struct {
    ubi_dlNode n;
    literal_t c[];
} cube_t;

cube_t*
create_cube(int degree)
{
    // maybe this should be a macro?
    return malloc(sizeof(cube_t) + sizeof(literal_t)*degree);
}

ubi_dlList*
create_set()
{
    ubi_dlList* set = malloc(sizeof(ubi_dlList));
    ubi_dlInitList(set);
    return set;
}

void
copycube(cube_t* src, cube_t* dest)
{
  int i;
  for (i=0;i<degree;i++)
    dest->c[i] = src->c[i];
}

ubi_dlListPtr
copy_set(ubi_dlListPtr set)
{
  ubi_dlListPtr s;
  ubi_dlNodePtr p;
  cube_t *c;
  
  s = create_set();
  for (p = ubi_dlFirst(set); p; p = ubi_dlNext(p))
  {
    c = create_cube( degree );
    copycube( (cube_t*)p, c );
    ubi_dlAddTail( s, (ubi_dlNodePtr)c );
  }
  return s;
}

void
free_set(ubi_dlListPtr set)
{
  if (set == NULL) return;
  while ( ubi_dlCount(set) )
    free( ubi_dlPop(set) );
  free( set );
}

// Need 6 mem locations for table -- want 3 zeros at the end. Any pointer
// ending in 0x0 or 0x8 will do.
unsigned char* table;
unsigned char  tabledata[] =
    { 0,      // 000 we should never hit this spot
      ZERO,   // 001 0*0 = 0
      ONE,    // 010 1*1 = 1
      EMPTY,  // 011 1*0 = 0
      X,      // 100 X*X = X
      ZERO,   // 101 X*0 = X
      ONE     // 110 X*1 = X 
     };

// (uchar, uchar) --> uchar
#define STAR(a,b)  (*((unsigned char*)(((unsigned int)table)|(a)|(b))))

#define NOT(x) (x == ONE ? ZERO : (x == ZERO ? ONE : EMPTY ))

void
init_star()
{
    // Do our magic to init the STAR operator.
    table = malloc(16);
    table += 8;
    table = (char*) (0xfffffff8 & (unsigned int) table);
    memcpy(table, tabledata, 7);
}

// Returns 1 if A is covered by B
int
covers(cube_t* a, cube_t* b)
{
    int i;
    for (i=0;i<degree;i++)
        if ( (a->c[i] != b->c[i]) && (b->c[i] != X))
            return 0;
    return 1;
}

// C = A * B
void
star(cube_t* a, cube_t* b, cube_t* c)
{
    int i, nulls = 0;
    for (i=0; i<degree; i++)
    {
        c->c[i] = STAR(a->c[i],b->c[i]);

        if (c->c[i] == EMPTY)
            nulls++;

        if (nulls > 1) 
        {
            c->c[0] = 0;
            return;
        }
        else
            c->c[i] = X;
    }
}

int eqcube(cube_t *a, cube_t *b)
{
  int i;
  for (i=0;i<degree;i++)
    if (a->c[i] != b->c[i])
      return 0;
  return 1;
}

/* equal assumes stable accumulation of cubes
 * existing cubes maintain their order
 */
int
equal(ubi_dlListPtr a, ubi_dlListPtr b)
{
  ubi_dlNodePtr pa,pb;
  if (ubi_dlCount(a) != ubi_dlCount(b))
    return 0;
  pa = ubi_dlFirst(a);
  pb = ubi_dlFirst(b);
  while ( (pa != NULL) && (pb != NULL) )
  {
    if ( !eqcube( (cube_t*)pa, (cube_t*)pb ) )
      return 0;
    pa = ubi_dlNext(pa);
    pb = ubi_dlNext(pb);
  }
  return 1;
}

ubi_dlListPtr
pi(ubi_dlListPtr co) //as in C naught
{
  ubi_dlListPtr g = NULL, c = NULL;
  while ( !equal(c,co) )
  {
    if (g)
      free_set(g);
    if (c)
      free_set(c);
    c = copy_set(co);
  }
  
}

void
nullify(cube_t* trgt)
{
  int i;
  for (i=0;i<degree;i++)
    trgt->c[i] = EMPTY;
}

void
sharp(cube_t* a, cube_t* b, cube_t* c)
{
  int nulls =0, eps=0, i=0;

  for (i=0;i<degree;i++)
  {
    if ((b->c[i] == X) || (b->c[i] == a->c[i]))
    { c->c[i] = EPSILON; eps++; } // do you like this style?
                                  // yuck!
    else 
    if (a->c[i] == X) 
    { c->c[i] = NOT(b->c[i]); }
    else
    { c->c[i] = EMPTY; nulls++; break; }
  }

  if (nulls)
    copycube(a,c);
  else if (eps == i)
    nullify(c);
  else
  {
    copycube(a,c);
    for (i=0;i<degree;i++)
      if ( (a->c[i] == X) && (b->c[i] != X) )
        c->c[i] = NOT(b->c[i]);
  }
  
}

// This is nice. too bad it's not needed anymore.
int 
factorial(int x)
{
    int n = 1;
    while (x) n*=x--;
    return n;
}

int
choose(int n, int k)
{
    return factorial(n)/factorial(k)/factorial(n-k);
}

literal_t
to_literal(char a)
{
    if (a == '0') return ZERO;
    if (a == '1') return ONE;
    if (a == 'x') return X;
    return 0;
}

char
from_literal(literal_t x)
{
    if (x == ZERO) return '0';
    if (x == ONE) return '1';
    if (x == X) return 'x';
    return 0;
}

// Has a terrible side effect - it sets 'degree'
ubi_dlList*
parse_blif(char* file)
{
    int n=-1, i=0, c, j;
    FILE* f;
    ubi_dlList* cubes = create_set();

    if ((f = fopen((const char*)file, "r")) == NULL) {
        perror("fopen");
        exit(1);
    }

    // count number of cubes
    while ((c = fgetc(f)) != EOF)
        if (c == (int) '\n') n++;
    rewind(f);

    // get past first line
    while ((c = fgetc(f)) != (int) '\n');

    // count degree
    while ((c = fgetc(f)) != (int) ' ') i++;
    fseek(f, -i-1, SEEK_CUR);

    // Setting globals for the good of humanity
    degree = i;
    printf("Cube of degree %d loaded.\n",degree);

    for (i=0; i<n; i++)
    {
        cube_t* cube = (cube_t*) ubi_dlPush(cubes, create_cube(degree));
        for (j=0; j<degree; j++)
            cube->c[j] = to_literal(fgetc(f));
        fgetc(f); // space
        fgetc(f); // assume it's a 1
        fgetc(f); // newline
    }

    return cubes;
}

void
print_set(ubi_dlList* s)
{
}


int
main(int argc, char** argv)
{
    ubi_dlList* bob;

    bob = parse_blif("blif/test1.blif");

    init_star();

    print_set(bob);

    return 0;
}
